import hy
import os
from .parse import parse_src, parse_node
from . import output, builtins
from pathlib import Path
from .utils import next_node_id_maker, build_node, next_nodeid


def reset_nodeid_counter():
    builtins.next_nodeid = next_node_id_maker()


def install_builtin_macros():
    macro_file = Path(os.path.dirname(__file__)).parent / "builtin" / "macros.hy"
    with macro_file.open() as f:
        code = f.read()
        for expr in hy.read_many(code):
            parse_node(expr)


install_builtin_macros()



from typing import List, Union
from hy import models
from dasy import parser
import vyper.ast.nodes as vy_nodes
from hy.models import Expression, Symbol

COMP_FUNCS = ["<", "<=", ">", ">=", "==", "!="]


def chain_comps(chain_expr: Expression) -> Expression:
    """
    Creates a new expression chaining comparisons.
    """
    new_node = models.Expression()
    new_expr: List[Union[Symbol, Expression]] = [models.Symbol("and")]
    for vals in zip(chain_expr[1:], chain_expr[2:]):
        new_expr.append(models.Expression((chain_expr[0], vals[0], vals[1])))
    new_node += tuple(new_expr)
    return new_node


def parse_comparison(comparison_expr: Expression) -> vy_nodes.Compare:
    """
    Parses a comparison expression, chaining comparisons if necessary.
    """
    assert (
        str(comparison_expr[0]) in COMP_FUNCS
    ), f"Invalid comparison operator {comparison_expr[0]}"

    # Always apply chain comps for consistency
    chained_expr = chain_comps(comparison_expr)
    left = parser.parse_node(chained_expr[1])
    right = parser.parse_node(chained_expr[2])
    op = parser.parse_node(chained_expr[0])
    return parser.build_node(vy_nodes.Compare, left=left, ops=[op], comparators=[right])



import ast as py_ast

from typing import Union

from dasy.parser.macros import handle_macro, is_macro


import hy
import vyper.ast.nodes as vy_nodes
from hy import models

from .builtins import parse_builtin, build_node
from .ops import BIN_FUNCS, BOOL_OPS, COMP_FUNCS, UNARY_OPS, is_op, parse_op
from .utils import add_src_map

# namespaces with expression handlers
from . import nodes, core, macros
from dasy.builtin import functions

BUILTIN_FUNCS = BIN_FUNCS | COMP_FUNCS | UNARY_OPS | BOOL_OPS | {"in", "notin"}

NAME_CONSTS = ["True", "False"]

CONSTS = {}

ALIASES = {
    ".": "attribute",
    "quote": "tuple",
    "array": "subscript",
    "defvar": "annassign",
    "setv": "assign",
    "+!": "unsafe_add",
    "-!": "unsafe_sub",
    "*!": "unsafe_mul",
    "/!": "unsafe_div",
    "def": "annassign",
}

SRC = ""


def convert_annassign(ast):
    # top-level AnnAssign nodes should be replaced with a VariableDecl
    is_public = False
    is_immutable = False
    is_constant = False
    if isinstance(ast.annotation, vy_nodes.Call):
        match ast.annotation.func:
            case "public":
                is_public = True
            case "immutable":
                is_immutable = True
            case "constant":
                is_constant = True
    new_node = build_node(
        vy_nodes.VariableDecl,
        target=ast.target,
        annotation=ast.annotation,
        value=ast.value,
        is_constant=is_constant,
        is_public=is_public,
        is_immutable=is_immutable,
    )
    for child in ast.get_children():
        new_node._children.add(child)
        child._parent = new_node
    return new_node


DONT_REPLACE = ("tuple",)


def parse_expr(expr):
    cmd_str = ALIASES.get(str(expr[0]), str(expr[0]))

    if cmd_str != str(expr[0]) and cmd_str not in DONT_REPLACE:
        expr = models.Expression((models.Symbol(cmd_str), *expr[1:]))

    if is_op(cmd_str):
        return parse_op(expr, cmd_str)

    if cmd_str in nodes.handlers:
        return nodes.handlers[cmd_str](expr)

    node_fn = f"parse_{cmd_str}"

    for ns in [nodes, core, macros, functions]:
        if hasattr(ns, node_fn):
            return getattr(ns, node_fn)(expr)

    if is_macro(cmd_str):
        return handle_macro(expr)

    if cmd_str.startswith("."):
        inner_node = models.Expression((models.Symbol("."), expr[1], (cmd_str[1:])))
        outer_node = models.Expression((inner_node, *expr[2:]))
        return parse_node(outer_node)

    match cmd_str:
        case "defconst":
            CONSTS[str(expr[1])] = expr[2]
            return None
        case "defimmutable" | "defimm":
            CONSTS[str(expr[1])] = None
            return None
        case "+=" | "-=" | "*=" | "/=":
            return parse_augop(expr)
        case _:
            return parse_call(expr)


def parse_augop(expr):
    op = models.Symbol(str(expr[0])[:1])
    target, value = expr[1:]
    parsed_code = build_node(
        vy_nodes.AugAssign,
        op=parse_node(op),
        target=parse_node(target),
        value=parse_node(value),
    )
    return parsed_code


def parse_call(expr, wrap_expr=False):
    match expr:
        case (fn_name, *args):
            args_list = []
            kw_args = []
            i = 0
            while i < len(args):
                cur_arg = args[i]
                if (
                    isinstance(cur_arg, models.Keyword)
                    and len(args) > (i + 1)
                    and not isinstance(args[i + 1], models.Keyword)
                ):
                    # TODO: remove this ugly hack and properly check against builtin types
                    # or reconsider whether we should be using keywords for builtin types at all
                    val_arg = args[i + 1]
                    val_node = parse_node(val_arg)
                    kw_node = build_node(
                        vy_nodes.keyword, arg=str(cur_arg)[1:], value=val_node
                    )
                    kw_args.append(kw_node)
                    i += 2
                else:
                    val_node = parse_node(args[i])
                    args_list.append(val_node)
                    i += 1
            func_node = parse_node(fn_name)
            call_node = build_node(
                vy_nodes.Call, func=func_node, args=args_list, keywords=kw_args
            )
            if wrap_expr:
                expr_node = build_node(vy_nodes.Expr, value=call_node)
                return expr_node
            return call_node


def parse_node(
    node: Union[
        models.Expression,
        models.Integer,
        models.String,
        models.Symbol,
        models.Keyword,
        models.Bytes,
        models.List,
    ],
):
    """
    This function converts a node into its corresponding AST node based on its type.
    :param node: A node of the parsed model
    :return: Corresponding AST node, if the node type is supported. Raises exception otherwise.
    """
    # Initialize ast_node to None
    ast_node = None

    # dispatch on type of parsed model
    match node:
        case models.Expression(node):
            ast_node = parse_expr(node)
        case models.Integer(node):
            ast_node = build_node(vy_nodes.Int, value=int(node))
        case models.String(node):
            ast_node = build_node(vy_nodes.Str, value=str(node))
        case models.Symbol(node):
            str_node = str(node)
            str_node = ALIASES.get(str_node, str_node)
            if str_node in CONSTS:
                ast_node = parse_node(CONSTS[str(node)])
            elif str_node in BUILTIN_FUNCS:
                ast_node = parse_builtin(node)
            elif str_node in NAME_CONSTS:
                ast_node = build_node(
                    vy_nodes.NameConstant, value=py_ast.literal_eval(str(node))
                )
            elif str_node.startswith("0x"):
                ast_node = build_node(vy_nodes.Hex, value=str_node)
            elif "/" in str_node:
                target, attr = str(node).split("/")
                replacement_node = models.Expression(
                    (models.Symbol("."), models.Symbol(target), models.Symbol(attr))
                )
                ast_node = parse_node(replacement_node)
            elif "." in str_node and len(str_node) > 1:
                target, attr = str(node).split(".")
                replacement_node = models.Expression(
                    (models.Symbol("."), models.Symbol(target), models.Symbol(attr))
                )
                ast_node = parse_node(replacement_node)
            else:
                ast_node = build_node(vy_nodes.Name, id=str(node))
        case models.Keyword(node):
            ast_node = build_node(vy_nodes.Name, id=str(node))
        case models.Bytes(byt):
            ast_node = build_node(vy_nodes.Bytes, value=byt)
        case models.List(lst):
            ast_node = build_node(
                vy_nodes.List, elements=[parse_node(elmt) for elmt in lst]
            )
        case models.Float(node):
            raise NotImplementedError("Floating point not supported (yet)")
        case _:
            raise ValueError(f"No match for node {node}. Unsupported node type.")
    return add_src_map(SRC, node, ast_node)


def parse_src(src: str):
    global SRC
    SRC = src
    mod_node: vy_nodes.Module = build_node(
        vy_nodes.Module, body=[], name="", doc_string=""
    )

    vars = []
    fs = []
    for element in hy.read_many(src):
        # parse each top-level form
        ast = parse_node(element)

        if isinstance(ast, list):
            for v in ast:
                add_src_map(src, element, v)
        elif ast:
            add_src_map(src, element, ast)
        else:
            continue

        if isinstance(ast, vy_nodes.Module):
            mod_node = ast
        elif isinstance(
            ast,
            (
                vy_nodes.VariableDecl,
                vy_nodes.StructDef,
                vy_nodes.EventDef,
                vy_nodes.InterfaceDef,
                vy_nodes.EnumDef,
            ),
        ):
            vars.append(ast)
        elif isinstance(ast, vy_nodes.FunctionDef):
            fs.append(ast)
        elif isinstance(ast, list):
            for var in ast:
                var_node = var
                if isinstance(var, vy_nodes.AnnAssign):
                    var_node = convert_annassign(var)
                elif isinstance(var, vy_nodes.FunctionDef):
                    fs.append(var)
                    continue
                vars.append(var_node)
        elif isinstance(ast, vy_nodes.AnnAssign):
            new_node = convert_annassign(ast)
            vars.append(new_node)
        else:
            raise Exception(f"Unrecognized top-level form {element} {ast}")

    for e in vars + fs:
        mod_node.add_to_body(e)

    return mod_node



import dasy
import vyper.ast.nodes as vy_nodes
from .utils import build_node, next_nodeid, pairwise
from hy import models

from .utils import has_return, process_body


def parse_attribute(expr):
    """Parses an attribute and builds a node."""
    if len(expr) < 2:
        raise ValueError("Expression too short to parse attribute.")
    _, obj, attr = expr
    attr_node = build_node(
        vy_nodes.Attribute, attr=str(attr), value=dasy.parser.parse_node(obj)
    )
    return attr_node


def parse_tuple(tuple_tree):
    elements = []
    if tuple_tree[0] == models.Symbol("quote"):
        elements = tuple_tree[1]
    elif tuple_tree[0] == models.Symbol("tuple"):
        elements = tuple_tree[1:]
    else:
        raise Exception("Invalid tuple declaration")
    return build_node(
        vy_nodes.Tuple, elements=[dasy.parser.parse_node(e) for e in elements]
    )


def parse_args_list(args_list) -> list[vy_nodes.arg]:
    if len(args_list) == 0:
        return []
    results = []
    current_type = args_list[0]
    assert isinstance(current_type, models.Keyword) or isinstance(
        current_type, models.Expression
    )
    # get annotation and name
    for arg in args_list[1:]:
        # check if we hit a new type
        if isinstance(arg, (models.Keyword, models.Expression)):
            current_type = arg
            continue
        # get annotation and name
        if isinstance(current_type, models.Keyword):
            # built-in types like :uint256
            annotation_node = build_node(
                vy_nodes.Name, id=str(current_type.name), parent=None
            )
        elif isinstance(current_type, models.Expression):
            # user-defined types like Foo
            annotation_node = dasy.parse.parse_node(current_type)
        else:
            raise Exception("Invalid type annotation")
        arg_node = build_node(
            vy_nodes.arg, arg=str(arg), parent=None, annotation=annotation_node
        )
        results.append(arg_node)
    return results


def parse_fn_args(fn_tree):
    args_node, *rest = fn_tree[2:]
    args_list = parse_args_list(args_node)
    args = build_node(vy_nodes.arguments, args=args_list, defaults=list())
    return args, rest


def parse_fn_decorators(decs):
    if isinstance(decs, models.Keyword):
        return [build_node(vy_nodes.Name, id=str(decs.name))]
    elif isinstance(decs, models.List):
        return [dasy.parse.parse_node(d) for d in decs]
    return []


def parse_fn_body(body, wrap=False):
    fn_body = [dasy.parse.parse_node(body_node) for body_node in body[:-1]]
    if wrap and not has_return(body[-1]):
        value_node = dasy.parse.parse_node(body[-1])
        implicit_return_node = build_node(vy_nodes.Return, value=value_node)
        fn_body.append(implicit_return_node)
    else:
        fn_body.append(dasy.parse.parse_node(body[-1]))
    return process_body(fn_body)


def _fn_tree_has_return_type(fn_tree):
    # has return type
    # (defn name [args] :uint256 :external ...)
    # (defn name [args] :uint256 [:external :view] ...)
    fn_args = fn_tree[1:]
    fn_args_len = len(fn_args)
    return (
        fn_args_len > 3
        and isinstance(fn_args[0], models.Symbol)
        and isinstance(fn_args[1], models.List)
        and isinstance(fn_args[2], (models.Keyword, models.Expression, models.Symbol))
        and isinstance(fn_args[3], (models.Keyword, models.List))
    )


def _fn_tree_has_no_return_type(fn_tree):
    # no return type
    # (defn name [args] ...)
    fn_args = fn_tree[1:]
    fn_args_len = len(fn_args)
    return (
        fn_args_len > 2
        and isinstance(fn_args[0], models.Symbol)
        and isinstance(fn_args[1], models.List)
        and isinstance(fn_args[2], (models.Keyword, models.List))
    )


def _fn_is_constructor(fn_tree):
    return isinstance(fn_tree[1], models.Symbol) and str(fn_tree[1]) == "__init__"


def parse_defn(fn_tree):
    fn_node_id = (
        next_nodeid()
    )  # we want our fn node to have a lower id than its child node
    assert isinstance(fn_tree, models.Expression)
    assert fn_tree[0] == models.Symbol("defn")
    return_type = None
    name = str(fn_tree[1])
    args = None
    decorators = []

    fn_args = fn_tree[1:]
    args, rest = parse_fn_args(fn_tree)

    if _fn_is_constructor(fn_tree):
        decorators = [build_node(vy_nodes.Name, id="external")]
        fn_body = parse_fn_body(rest[1:])
    elif _fn_tree_has_return_type(fn_tree):
        decorators = parse_fn_decorators(fn_args[3])
        fn_body = parse_fn_body(rest[2:], wrap=True)
        return_type = dasy.parse.parse_node(fn_args[2])
    elif _fn_tree_has_no_return_type(fn_tree):
        decorators = parse_fn_decorators(fn_args[2])
        fn_body = parse_fn_body(rest[1:])
    else:
        raise Exception(f"Invalid fn form {fn_tree}")

    fn_node = build_node(
        vy_nodes.FunctionDef,
        args=args,
        returns=return_type,
        decorator_list=decorators,
        pos=None,
        body=fn_body,
        name=name,
        node_id=fn_node_id,
    )

    return fn_node


def parse_declaration(var, typ, value=None):
    target = dasy.parse.parse_node(var)
    annotation_attrs = {"public": False, "immutable": False, "constant": False}
    annotation = None

    match typ:
        case [models.Symbol(e), _] if str(e) in annotation_attrs.keys():
            annotation = dasy.parse.parse_node(typ)
            annotation_attrs[str(e)] = True
        case models.Expression() | models.Keyword():
            annotation = dasy.parse.parse_node(typ)
        case _:
            raise Exception(f"Invalid declaration type {typ}")

    if annotation is None:
        raise Exception("No valid annotation was found")

    vdecl_node = build_node(
        vy_nodes.VariableDecl,
        target=target,
        annotation=annotation,
        value=value,
        **annotation_attrs,
    )
    return vdecl_node


def parse_defvars(expr):
    return [parse_declaration(var, typ) for var, typ in pairwise(expr[1:])]


def create_annotated_node(node_class, var, typ, value=None):
    target = dasy.parse.parse_node(var)
    if not isinstance(typ, (models.Expression, models.Keyword, models.Symbol)):
        raise Exception(f"Invalid declaration type {typ}")
    annotation = dasy.parse.parse_node(typ)
    node = build_node(node_class, target=target, annotation=annotation, value=value)
    return node


def parse_variabledecl(expr) -> vy_nodes.VariableDecl:
    return create_annotated_node(
        vy_nodes.VariableDecl,
        expr[1],
        expr[2],
        value=dasy.parse.parse_node(expr[3]) if len(expr) == 4 else None,
    )


def parse_annassign(expr) -> vy_nodes.AnnAssign:
    return create_annotated_node(
        vy_nodes.AnnAssign,
        expr[1],
        expr[2],
        value=dasy.parse.parse_node(expr[3]) if len(expr) == 4 else None,
    )


def parse_structbody(expr):
    return [
        create_annotated_node(vy_nodes.AnnAssign, var, typ)
        for var, typ in pairwise(expr[2:])
    ]


def parse_defcontract(expr):
    mod_node = vy_nodes.Module(
        body=[],
        name=str(expr[1]),
        doc_string="",
        ast_type="Module",
        node_id=next_nodeid(),
    )
    expr_body = []
    match expr[1:]:
        case (_, vars, *body) if isinstance(vars, models.List):
            # # contract has state
            for var, typ in pairwise(vars):
                mod_node.add_to_body(parse_declaration(var, typ))
            expr_body = body
        case (_, *body):
            # no contract state
            expr_body = body
        case _:
            raise Exception(f"Invalid defcontract form: {expr}")
    for node in expr_body:
        mod_node.add_to_body(dasy.parse.parse_node(node))

    return mod_node


def parse_defstruct(expr):
    struct_node = build_node(
        vy_nodes.StructDef, name=str(expr[1]), body=parse_structbody(expr)
    )
    return struct_node


def parse_definterface(expr):
    name = str(expr[1])
    body = []
    for f in expr[2:]:
        rets = None if len(f) == 4 else dasy.parse.parse_node(f[3])

        args_list = parse_args_list(f[2])
        args_node = build_node(vy_nodes.arguments, args=args_list, defaults=list())

        # in an interface, the body is a single expr node with the visibility
        visibility_node = dasy.parse.parse_node(f[-1])
        body_node = build_node(vy_nodes.Expr, value=visibility_node)

        fn_node = build_node(
            vy_nodes.FunctionDef,
            args=args_node,
            returns=rets,
            decorator_list=[],
            pos=None,
            body=[body_node],
            name=str(f[1]),
        )
        body.append(fn_node)

    interface_node = build_node(vy_nodes.InterfaceDef, body=body, name=name)
    return interface_node


def parse_defevent(expr):
    return build_node(vy_nodes.EventDef, name=str(expr[1]), body=parse_structbody(expr))


def parse_enumbody(expr):
    return [build_node(vy_nodes.Expr, value=dasy.parse.parse_node(x)) for x in expr[2:]]


def parse_defenum(expr):
    return build_node(vy_nodes.EnumDef, name=str(expr[1]), body=parse_enumbody(expr))


def parse_do(expr):
    calls = [dasy.parse.parse_node(x) for x in expr[1:]]
    exprs = [build_node(vy_nodes.Expr, value=call_node) for call_node in calls]
    return exprs


def parse_subscript(expr):
    """(subscript value slice)"""
    index_value_node = dasy.parse.parse_node(expr[2])
    index_node = build_node(vy_nodes.Index, value=index_value_node)
    value_node = dasy.parse.parse_node(expr[1])
    subscript_node = build_node(vy_nodes.Subscript, slice=index_node, value=value_node)
    return subscript_node



# for handling venom
from vyper.ast import Call, Expr
from vyper.ir.s_expressions import parse_s_exp
from vyper.codegen.ir_node import IRnode
from vyper.builtins.functions import STMT_DISPATCH_TABLE, DISPATCH_TABLE, BuiltinFunction
from vyper.compiler import phases
from vyper.semantics.types.shortcuts import UINT256_T
from vyper.codegen.types._types import parse_type
from vyper.semantics.types.utils import type_from_abi

from dasy import parser
from dasy.parser.utils import get_ir_type

from hy import repr


def parse_ir(expr):
    # check for optional return type annotation as second element
    ret_type = None
    ir_expr = None
    if len(expr) == 3:
        ret_type = get_ir_type(expr[1].name)
        ir_expr = expr[2]
    elif len(expr) == 2:
        ir_expr = expr[1]
    ir = IRnode.from_list((parse_s_exp(repr(ir_expr)[1:]))[0], typ=ret_type)

    # generate some vyper code to patch in.
    IDENTIFIER = f"__DASY_VENOM_BUILTIN_{parser.next_nodeid()}__"
    insert_code = f"{IDENTIFIER}()"

    # dynamically generate a class on the fly
    class generated_builtin(BuiltinFunction):
        _id = IDENTIFIER
        _inputs = ()
        _return_type = ret_type

        def fetch_call_return(self, node):
            return self._return_type

        def infer_arg_types(self, node):
            return []

        def build_IR(self, expr, context):
            return ir

    if ret_type is not None:
        DISPATCH_TABLE[IDENTIFIER] = generated_builtin()
        gend_ast = phases.generate_ast(insert_code, 0, "")
        return gend_ast[1].body[0].value

    STMT_DISPATCH_TABLE[IDENTIFIER] = generated_builtin()
    return phases.generate_ast(insert_code, 0, "")[1].body[0]


def parse_vyper(expr):
    return phases.generate_ast(str(expr[1]), 0, "")[1].body[0]


def wrap_calls(nodes):
    new_nodes = []
    for call_node in nodes:
        if isinstance(call_node, Call):
            expr_node = parser.build_node(Expr, value=call_node)
            new_nodes.append(expr_node)
        else:
            new_nodes.append(call_node)
    return new_nodes


def parse_splice(expr):
    return_val = wrap_calls([parser.parse_node(n) for n in expr[1:]])
    return return_val



from typing import List, Union
from hy import models
from dasy import parser
from vyper.ast.nodes import BinOp, Compare, UnaryOp, BoolOp
from .builtins import build_node

BIN_FUNCS = {"+", "-", "/", "*", "**", "%"}
COMP_FUNCS = {"<", "<=", ">", ">=", "==", "!=", "in", "notin"}
UNARY_OPS = {"not", "usub"}
BOOL_OPS = {"and", "or"}


def is_op(cmd_str):
    return cmd_str in BIN_FUNCS | COMP_FUNCS | UNARY_OPS | BOOL_OPS


def parse_op(expr, alias=None):
    cmd_str = alias or str(expr[0])
    if cmd_str in BIN_FUNCS:
        return parse_binop(expr)
    if cmd_str in COMP_FUNCS:
        return parse_comparison(expr)
    if cmd_str in UNARY_OPS:
        return parse_unary(expr)
    if cmd_str in BOOL_OPS:
        return parse_boolop(expr)


def chain_comps(expr):
    new_node = models.Expression()
    new_expr: List[Union[models.Symbol, models.Expression]] = [models.Symbol("and")]
    for vals in zip(expr[1:], expr[2:]):
        new_expr.append(models.Expression((expr[0], vals[0], vals[1])))
    new_node += tuple(new_expr)
    return new_node


def parse_comparison(comp_tree):
    if (
        len(comp_tree[1:]) > 2
    ):  # comparing more than 2 things; chain comps for (< 2 3 4 )
        return parser.parse_node(chain_comps(comp_tree))
    left = parser.parse_node(comp_tree[1])
    right = parser.parse_node(comp_tree[2])
    op = parser.parse_node(comp_tree[0])
    return build_node(Compare, left=left, ops=[op], comparators=[right])


def parse_unary(expr):
    operand = parser.parse_node(expr[1])
    op = parser.parse_node(expr[0])
    return build_node(UnaryOp, operand=operand, op=op)


def parse_boolop(expr):
    op = parser.parse_node(expr[0])
    values = [parser.parse_node(e) for e in expr[1:]]
    return build_node(BoolOp, op=op, values=values)


def chain_binops(expr):
    if len(expr) == 3:
        return expr
    else:
        new_node = models.Expression()
        tmp_expr = tuple([expr[0], *expr[2:]])
        tmp_node = models.Expression()
        tmp_node += tmp_expr
        subtree = chain_binops(tmp_node)
        new_node += tuple([expr[0], expr[1], subtree])
        return new_node


def parse_binop(binop_tree):
    if len(binop_tree) > 3:
        return parser.parse_node(chain_binops(binop_tree))
    left = parser.parse_node(binop_tree[1])
    right = parser.parse_node(binop_tree[2])
    op = parser.parse_node(binop_tree[0])
    return build_node(BinOp, left=left, right=right, op=op)



from dasy import compiler
from vyper.compiler import OUTPUT_FORMATS as VYPER_OUTPUT_FORMATS
import argparse
import sys

from dasy.parser.output import get_external_interface

format_help = """Format to print, one or more of:
bytecode (default) - Deployable bytecode
bytecode_runtime   - Bytecode at runtime
abi                - ABI in JSON format
abi_python         - ABI in python format
source_map         - Vyper source map
method_identifiers - Dictionary of method signature to method identifier
userdoc            - Natspec user documentation
devdoc             - Natspec developer documentation
combined_json      - All of the above format options combined as single JSON output
layout             - Storage layout of a Vyper contract
ast                - AST in JSON format
external_interface - External (Dasy) interface of a contract, used for outside contract calls
vyper_interface    - External (Vyper) interface of a contract, used for outside contract calls
opcodes            - List of opcodes as a string
opcodes_runtime    - List of runtime opcodes as a string
ir                 - Intermediate representation in list format
ir_json            - Intermediate representation in JSON format
hex-ir             - Output IR and assembly constants in hex instead of decimal
no-optimize        - Do not optimize (don't use this for production code)
"""

OUTPUT_FORMATS = VYPER_OUTPUT_FORMATS.copy()

OUTPUT_FORMATS["vyper_interface"] = OUTPUT_FORMATS["external_interface"]
OUTPUT_FORMATS["external_interface"] = get_external_interface


def main():
    parser = argparse.ArgumentParser(
        prog="dasy",
        description="Lispy Smart Contract Language for the EVM",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("filename", type=str, nargs="?", default="")
    parser.add_argument("-f", help=format_help, default="bytecode", dest="format")

    src = ""

    args = parser.parse_args()

    if args.filename != "":
        with open(args.filename, "r") as f:
            src = f.read()
            if args.filename.endswith(".vy"):
                data = compiler.CompilerData(
                    src, contract_name=args.filename.split("/")[-1].split(".")[0]
                )
            else:
                data = compiler.compile(src, name=args.filename.split(".")[0])
    else:
        for line in sys.stdin:
            src += line
        data = compiler.compile(src, name="StdIn")

    translate_map = {
        "abi_python": "abi",
        "json": "abi",
        "ast": "ast_dict",
        "ir_json": "ir_dict",
        "interface": "external_interface",
    }
    output_format = translate_map.get(args.format, args.format)
    if output_format in OUTPUT_FORMATS:
        print(OUTPUT_FORMATS[output_format](data))
    else:
        raise Exception(
            f"Unrecognized Output Format {args.format}. Must be one of {OUTPUT_FORMATS.keys()}"
        )


if __name__ == "__main__":
    main()



from vyper.ast import nodes as vy_nodes
from vyper.ast.nodes import (
    Break,
    Pass,
    Continue,
    Log,
    Raise,
    Return,
    AugAssign,
    Assert,
    Index,
)
from hy import models
from dasy import parser
from .utils import process_body, build_node


def parse_for(expr):
    # (for [x xs] (.append self/nums x))
    # (for [target iter] *body)
    target, iter_ = expr[1]
    target_node = parser.parse_node(target)
    iter_node = parser.parse_node(iter_)
    body_nodes = [parser.parse_node(b) for b in expr[2:]]
    body = process_body(body_nodes)
    for_node = build_node(vy_nodes.For, body=body, iter=iter_node, target=target_node)
    return for_node


def parse_if(expr):
    # used for base case in cond expansion
    if expr[1] == models.Keyword("else"):
        if expr[3] == models.Symbol("None"):
            return parser.parse_node(expr[2])

    body_nodes = [parser.parse_node(expr[2])]
    body = process_body(body_nodes)
    else_nodes = [parser.parse_node(expr[3])] if len(expr) == 4 else []
    else_ = process_body(else_nodes)
    test = parser.parse_node(expr[1])

    # if-expressions always have:
    # - one node in body
    # - one node in else
    # - both nodes are ExprNodes
    # in theory we could also verify that both ExprNodes are of the same type
    # but the Vyper compiler will catch that anyway
    if (
        len(body) == 1
        and len(else_) == 1
        and isinstance(body[0], vy_nodes.ExprNode)
        and isinstance(else_[0], vy_nodes.ExprNode)
    ):
        body = body[0]
        else_ = else_[0]
        if_node = build_node(vy_nodes.IfExp, test=test, body=body, orelse=else_)
    else:
        if_node = build_node(vy_nodes.If, test=test, body=body, orelse=else_)
    return if_node


def parse_assign(expr):
    # needs some slight massaging due to the way targets/target is treated
    # the Assign class has a target slot, but it uses the first value in the
    # targets arg to fill it instead of using the target kwarg
    args = [parser.parse_node(arg) for arg in expr[1:]]
    return build_node(vy_nodes.Assign, *args, targets=[args[0]])


def parse_expr(expr, nodes):
    return [parser.parse_node(node) for node in expr[1 : nodes + 1]]


handlers = {
    node_type.__name__.lower(): lambda expr, node_type=node_type: build_node(
        node_type, *parse_expr(expr, 2)
    )
    for node_type in [
        Break,
        Pass,
        Continue,
        Log,
        Raise,
        Return,
        AugAssign,
        Assert,
        Index,
    ]
}



import dasy
import hy

MACROS = ["cond"]


def is_macro(cmd_str):
    return cmd_str in MACROS


def macroexpand(code_str):
    return hy.macroexpand(hy.read(code_str))


def handle_macro(expr):
    new_node = hy.macroexpand(expr)
    return dasy.parser.parse_node(new_node)


def parse_defmacro(expr):
    hy.eval(expr)
    MACROS.append(str(expr[1]))
    return None



from vyper.compiler.phases import CompilerData as VyperCompilerData
from pathlib import Path
from vyper.compiler.output import (
    build_abi_output,
    build_asm_output,
    build_bytecode_runtime_output,
    build_external_interface_output,
    build_interface_output,
    build_ir_output,
    build_ir_runtime_output,
    build_layout_output,
    build_opcodes_output,
)
from dasy.parser import parse_src
from dasy.parser.utils import filename_to_contract_name


class CompilerData(VyperCompilerData):
    def __init__(self, *args, **kwargs):
        VyperCompilerData.__init__(self, *args, **kwargs)

    @property
    def runtime_bytecode(self):
        runtime_bytecode = build_bytecode_runtime_output(self)
        self.__dict__["runtime_bytecode"] = runtime_bytecode
        return runtime_bytecode

    @property
    def abi(self):
        abi = build_abi_output(self)
        self.__dict__["abi"] = abi
        return abi

    @property
    def interface(self):
        interface = build_interface_output(self)
        self.__dict__["interface"] = interface
        return interface

    @property
    def ir(self):
        ir = build_ir_output(self)
        self.__dict__["ir"] = ir
        return ir

    @property
    def runtime_ir(self):
        ir = build_ir_runtime_output(self)
        self.__dict__["runtime_ir"] = ir
        return ir

    @property
    def asm(self):
        asm = build_asm_output(self)
        self.__dict__["asm"] = asm
        return asm

    @property
    def opcodes(self):
        return build_opcodes_output(self)

    @property
    def runtime_opcodes(self):
        return build_opcodes_output(self)

    @property
    def external_interface(self):
        return build_external_interface_output(self)

    @property
    def layout(self):
        return build_layout_output(self)


def generate_compiler_data(src: str, name="DasyContract") -> CompilerData:
    ast = parse_src(src)
    data = CompilerData(
        "",
        ast.name or name,
        None,
        source_id=0,
    )
    data.vyper_module = ast
    return data


def compile(src: str, name="DasyContract", include_abi=True) -> CompilerData:
    data = generate_compiler_data(src, name)
    return data


def compile_file(filepath: str) -> CompilerData:
    path = Path(filepath)
    name = path.stem
    # name = ''.join(x.capitalize() for x in name.split('_'))
    with path.open() as f:
        src = f.read()
        if filepath.endswith(".vy"):
            return CompilerData(src, contract_name=filename_to_contract_name(filepath))
        return compile(src, name=name)


def generate_abi(src: str) -> list:
    return compile(src).abi
